%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Son Tran's translator program - tr.pl
%%
%% Input:  a Prolog program representing a domain description
%%         in Language A_K with sensing actions.
%%
%% Output: a Smodels program equivalent to the input program.
%%
%% This translator accepts A_k programs written in the following
%% syntax:
%% 1. Action Declarations
%%    action(A).        (A is an action)
%%
%% 2. Fluent Declarations
%%    fluent(F).        (F is a fluent)
%%
%% 3. Executability Conditions:
%%    executable(A,Ls).  (A can be executed iff Ls holds)
%%
%% 4. Dynamic Causal Laws:
%%    causes(A,L,Ls).    (A causes literal L
%%		                     to hold if C holds)
%%
%% 5. Sensing Actions' Effects:
%%    determines(A,Ls). (A determines literals Ls)
%%
%% 6. Static Causal Laws:
%%    caused(L,Ls).      (L holds in every state where
%%	                         condition Ls holds)
%%    
%% 7. Initial State:
%%    initially(L).     (literal L holds in the initial state)  
%%
%% 8. Goal state:
%%    goal(L).          (literal L holds in the final state)
%%
%% 9. Abbreviations
%%    oneof(Fs)         (fluents in Fs are mutually exclusive)
%% Note that Ls is a list of literals and literals L's are of
%% the form f or neg(f), where f is some constant. A is an action
%% and F is a fluent. Actions and fluents used in the program
%% must be declared explicitly using predicates action and fluent.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- use_module(library(lists)).
:- use_module(library(system)).
:- use_module(library(charsio)).

%:- dynamic (fluent/1).
%:- dynamic
%	initially/1,
%	goal/1,
%	fluent/1,
%	action/1,
%	executable/2,
%	causes/3,
%	caused/2,
%	determines/2,
%	oneof/1.

%:- multifile
%	initially/1,
%	goal/1,
%	causes/3,
%	caused/2,
%	determines/2,
%	oneof/1.

unknown_predicate_handler(_,_,fail).

:-prolog_flag(redefine_warnings,_,off). 

% the following flag is used for swi prolog only
% set_prolog_flag(unknown,fail).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Main procedure
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% the main procedure for generating the smodels program for
% a domain description written in A_K
main :-
	write('>> Enter domain description file name: '),
	read(S),
	name(S,Chars),
	name(smo,Chars1),
	append(Chars,[46|Chars1],Chars2),
	name(S1,Chars2),
	absolute_file_name(S,In),       % get the source file name
	absolute_file_name(S1,Out),	    % get the domain output file name
	gen(In,Out).

gen(In,Out) :-
	abolish(fluent/1),
	abolish(action/1),
	abolish(causes/3),
	abolish(caused/2),
	abolish(oneof/1),
	abolish(determines/2),
	abolish(initially/1),
	abolish(goal/1),
	abolish(executable/2),
	consult(In),                               % compile the source file
	open(Out,write,Id),                        % open file for writing
	write('Generating... \n'),
	findall(A1,action(A1),AL),	               % action list
	findall(F1,fluent(F1),FL),                 % fluent list
	findall((A2,C2),executable(A2,C2),ExL),	   % executability conditions
	findall((A3,L3,C3),causes(A3,L3,C3),EfL),  % dynamic laws
	findall((A4,F4),determines(A4,F4),SensEfL),% k-propositions
	findall((L5,C5),caused(L5,C5),StatL),      % static laws
	findall(Fs,oneof(Fs),OL),                  % oneof rules
	findall(L5,initially(L5),IL),              % intial state
	findall(L6,goal(L6),GL),                   % goal state
	findall(F7,determines(_,F7),SL),           % sensed fluents
	pr_header(Id),						       % print program usage
	pr_section(Id,'% Action declarations\n'),
	pr_actions(Id,AL),			% print action declarations
	pr_section(Id,'% Fluent declarations\n'),
	pr_fluents(Id,FL),			% print fluent declarations
	pr_sensed_literals(Id,SL,FL),% print sensed literals
	pr_section(Id,'% DOMAIN INDEPENDENT RULES \n'),
	pr_independent(Id),
	pr_section(Id,'% DOMAIN DEPENDENT RULES \n'),
	write(Id,'% Executability conditions \n'),
	pr_exec(Id,ExL,AL,FL),		    % print executability conditions
	nl(Id),
	write(Id,'% Static laws \n'),	
	pr_static_laws(Id,StatL,FL),	% print static laws
	pr_oneof(Id,OL,FL),			    % prine one-of rules
	nl(Id),
	write(Id,'% Effects of non-sensing actions \n'),		
	pr_effects(Id,EfL,AL,FL),		% print effects of non-sensing actions
	nl(Id),
	write(Id,'% Effects of sensing actions \n'),		
	pr_sense_effects(Id,SensEfL,AL,FL), % print effects of sensing actions
	pr_section(Id,'% INITIAL STATE \n'),
	pr_init(Id,IL,FL),
	pr_section(Id,'% GOAL SPECIFICATION \n'),
	pr_goal(Id,GL,FL),
	pr_section(Id,'% HIDE THE FOLLOWING ATOMS \n'),
	pr_hide(Id),
	close(Id), % close file
	write('done.\n').

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Common predicates
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
check_fluent(F,FL) :-
	(non_member(F,FL) -> write('\tError: fluent '), write(F),
		write(' is not declared.\n'), fail; true).

check_literal(L,FL) :-
	negate(L,L1),
	(non_member(L,FL), non_member(L1,FL)) ->
	write('\tError: literal '), write(L),
	write(' is not valid. Check fluent declarations.\n'), fail; true.

check_literals([],_).
check_literals([H|T],FL) :-
	check_literal(H,FL), check_literals(T,FL).

check_action(A,AL) :-
	(non_member(A,AL) -> write('\tError: action '), write(A),
		write(' is not declared.\n'), fail; true).

% negate(?L1,?L2): L1 & L2 negates each other
negate(neg(F),F) :- !.
negate(F,neg(F)) :- !.
negate_list([],[]).
negate_list([H1|T1],[H2|T2]) :-
	negate(H1,H2), negate_list(T1,T2).

% pr_line(+Id): print a comment line
pr_line(Id) :-
	write(Id,'%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%'),
	write(Id,'%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n').

% pr_section(+Id,+Cmt): create a new section
pr_section(Id,Cmt) :-
	nl(Id),
	pr_line(Id),
	write(Id,Cmt),
	pr_line(Id).

% pr_header(Id): print program usage
pr_header(Id) :-
	pr_line(Id),
	write(Id,'% Usage:                              \n'),
	write(Id,'%     lparse -c h=<height> -c w=<width> | smodels\n'),
	write(Id,'% The program consists of the following main predicates\n'),
	write(Id,'%\tholds(L,T,P) - if literal L holds at (T,P)\n'),
	write(Id,'%\tph(L,T,P) - if literal L possibly holds at (T,P)\n'),
	write(Id,'%\tposs(A,T,P) - if action A is executable at (T,P)\n'),
	write(Id,'%\tocc(A,T,P) - if action A occurs at (T,P)\n'),
	write(Id,'%\tused(T,P) - if starting from T, the path P is \n'),
	write(Id,'%\t\tused in constructing the plan.\n'),
	write(Id,'%\tbr(g,T,P,P_1) - if there exists a branch from (T,P)\n'),
	write(Id,'%\t\tto (T+1,P_1)\n'),
	write(Id,'%\tgoal(T,P) - if the goal is satisified at (T,P)\n'),
	write(Id,'%\tpc(L,T,P) - if L may change in the next state\n'),
	write(Id,'%\te(L,T,P) - if literal L is an effect of some action\n'),
	write(Id,'%\t\toccuring at (T,P).\n'),
	pr_line(Id),
	write(Id,'% Input parameters\n'),
	write(Id,'time(1..h).\n'),
	write(Id,'time1(1..h+1).\n'),
	write(Id,'path(1..w).\n').

% pr_rule(+Id,+Hd,+B): prints a rule with Head H and Body B
pr_rule(Id,Hd,[]) :-
	write(Id,Hd), write(Id,'.\n').

pr_rule(Id,Hd,[H|T]) :-
	(Hd \== '' ->
	(write(Id,Hd), write(Id,' :-\n\t'),
	write(Id,H), pr_list(Id,T));
	(write(Id,':- '), write(Id,H), pr_list(Id,T))).

pr_list(Id,[]) :-
	write(Id,'.\n').

pr_list(Id,[H|T]) :-
	write(Id,',\n\t'), write(Id,H),
	pr_list(Id,T).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Generate action description 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% pr_actions(+Id,+AL) : print action declarations
pr_actions(_,[]).
pr_actions(Id,[H|T]) :-
	write(Id,'action('), write(Id,H), write(Id,').\n'),
	pr_actions(Id,T).

% pr_fluents(+Id,+FL) : print fluent declarations
pr_fluents(_,[]).
pr_fluents(Id,[H|T]) :-
	write(Id,'fluent('), write(Id,H), write(Id,').\n'),
	pr_fluents(Id,T).

% pr_sensed_literals(+Id,+SL,+FL) : print sensed fluent declarations
pr_sensed_literals(_,[],_).
pr_sensed_literals(Id,[H|T],FL) :-
	pr_sensed_list(Id,H,FL),
	pr_sensed_literals(Id,T,FL).

pr_sensed_list(_,[],_).
pr_sensed_list(Id,[H|T],FL) :-
	check_literal(H,FL),
	write(Id,'sense('), write(Id,H), write(Id,').\n'),
	pr_sensed_list(Id,T,FL).

% pr_independent(Id): print domain independent rules
pr_independent(Id) :-
	write(Id,'% Auxiliary Rules\n'),
	pr_rule(Id,'literal(F)',['fluent(F)']),
	pr_rule(Id,'literal(neg(F))',['fluent(F)']),
	nl(Id),
	pr_rule(Id,'contrary(F,neg(F))',['fluent(F)']),
	pr_rule(Id,'contrary(neg(F),F)',['fluent(F)']),
	nl(Id),
	pr_rule(Id,'ph(L,T,P)',['literal(L)','time(T)',
		'path(P)','literal(L1)','contrary(L,L1)',
		'not holds(L1,T,P)']),
	nl(Id),
	pr_rule(Id,'used(1,1)',[]),
	pr_rule(Id,'used(T+1,P)',['time(T)','path(P)','used(T,P)']),
	nl(Id),
	write(Id,'% Goal representation\n'),
	pr_rule(Id,'goal(T+1,P)',['time(T)','path(P)',
		'goal(T,P)']),
	nl(Id),
	pr_rule(Id,'',['path(P)','used(h+1,P)','not goal(h+1,P)']),
	nl(Id),
	write(Id,'% Inertial rules for non-sensing actions\n'),
	pr_rule(Id,'holds(L,T+1,P)',['literal(L)','time(T)',
		'path(P)','e(L,T+1,P)']),
	nl(Id),
	pr_rule(Id,'holds(L,T+1,P)',['literal(L)','time(T)',
		'path(P)','holds(L,T,P)','literal(L1)','contrary(L,L1)',
		'not pc(L1,T+1,P)']),
	nl(Id),
	write(Id,'% Inertial rules for sensing actions\n'),
	pr_rule(Id,'new_br(P,P1)',['path(P)','path(P1)',
		'P < P1']),
	nl(Id),
	pr_rule(Id,'',['sense(G1)','sense(G2)','neq(G1,G2)',
		'time(T)','path(P)','path(P1)','path(P2)',
		'br(G1,T,P1,P)','br(G2,T,P2,P)']),
	nl(Id),
	pr_rule(Id,'',['sense(G1)','sense(G2)',
		'time(T)','path(P)','path(P1)','path(P2)',
		'neq(P1,P2)','br(G1,T,P1,P)','br(G2,T,P2,P)']),
	nl(Id),
	pr_rule(Id,'',['sense(G)','time(T)','path(P)',
		'path(P1)','neq(P1,P)','br(G,T,P1,P)','used(T,P)']),
	nl(Id),
	pr_rule(Id,'used(T+1,P)',['time(T)','path(P)',
		'path(P1)','sense(G)','br(G,T,P1,P)']),
	nl(Id),
	pr_rule(Id,'holds(G,T+1,P)',['sense(G)','time(T)',
		'path(P)','path(P1)','br(G,T,P1,P)']),
	nl(Id),
	pr_rule(Id,'holds(L,T+1,P)',['literal(L)','time(T)',
		'path(P)','path(P1)','sense(G)','br(G,T,P1,P)',
		'holds(L,T,P1)']),
	nl(Id),
	write(Id,'% Rules for generating action occurrences\n'),
	pr_rule(Id,'1{occ(A,T,P):action(A)}1',['time(T)',
		'path(P)','used(T,P)','not goal(T,P)']),
	nl(Id),
	pr_rule(Id,'',['action(A)','time(T)','path(P)',
		'occ(A,T,P)','not poss(A,T,P)']).

% pr_init(+Id,+IL,+FL): print rules encoding initial state
pr_init(_,[],_).
pr_init(Id,[H|T],FL) :-
	check_literal(H,FL),
	pr_rule(Id,holds(H,1,1),[]),
	pr_init(Id,T,FL).

% pr_goal(+Id,+GL,+FL): print rules specifying the goal
pr_goal(Id,GL,FL) :-
	check_literals(GL,FL),
	make_list(holds,',T,P',GL,NL),
	append(['time1(T)','path(P)'],NL,B),
	pr_rule(Id,'goal(T,P)',B).

% make_list(+Prefix,+Postfix,+InputList,?OutputAtomList)
make_list(_,_,[],[]).
make_list(Pr,Ps,[H|T],NL) :-
	make_atom(Pr,Ps,H,A),
	make_list(Pr,Ps,T,L),append([A],L,NL).

% make_atom(+Prefix,+Postfix,+Chars,?Atom)
make_atom(Pr,Ps,X,A) :-
	name(Pr,S1), name('(',S2), write_to_chars(X,S3),
	name(Ps,S4), name(')',S5), append(S1,S2,L1),
	append(L1,S3,L2), append(L2,S4,L3),
	append(L3,S5,L4), name(A,L4).

concat(S1,S2,S3) :-
	name(S1,L1), name(S2,L2), append(L1,L2,L3),
	name(S3,L3).

% pr_exec(+Id,+LC,+AL,+FL): print the executability conditions
pr_exec(_,[],_,_).
pr_exec(Id,[(A,C)|T],AL,FL) :-
	check_action(A,AL),
	check_literals(C,FL),
	make_atom(poss,',T,P',A,H),
	make_list(holds,',T,P',C,B),
	append(['time(T)','path(P)'],B,B1),
	pr_rule(Id,H,B1),
	pr_exec(Id,T,AL,FL).

% pr_static_laws(+Id,+StatL,+FluentList): print static laws
pr_static_laws(_,[],_).
pr_static_laws(Id,[(H,C)|T],FL) :-
	check_literal(H,FL),
	check_literals(C,FL),
	pr_static_law(Id,H,C),
	nl(Id),
	pr_static_laws(Id,T,FL).

% pr_static_law(+Id,+Head,+Body): print a static law
pr_static_law(Id,H,C) :-
	make_atom(holds,',T,P',H,Hd),
	make_list(holds,',T,P',C,B),
	append(['time1(T)','path(P)'],B,Bd),
	pr_rule(Id,Hd,Bd),
	nl(Id),
	make_atom(e,',T+1,P',H,Hd1),
	make_list(e,',T+1,P',C,B1),
	append(['time(T)','path(P)'],B1,Bd1),
	pr_rule(Id,Hd1,Bd1),
	nl(Id),
	pr_abnormality(Id,H,C,C).

%pr_abnormality(+Id,+Head,+Body,+Body)
pr_abnormality(_,_,[],_).
pr_abnormality(Id,Hd,[H|T],Bd) :-
	negate_list(Bd,Bd1),
	make_atom(pc,',T+1,P',Hd,Hd1),
	make_atom(pc,',T+1,P',H,B1),
	make_atom('not holds',',T,P',Hd,B2),
	make_list('not e',',T+1,P',Bd1,Bd2),
	append(['time(T)','path(P)',B1,B2],Bd2,Bd3),
	pr_rule(Id,Hd1,Bd3),
	pr_abnormality(Id,Hd,T,Bd).

% pr_oneof(+Id,+Rules,+FluentList): print static laws for one-of rules.
pr_oneof(_,[],_).
pr_oneof(Id,[H|T],FL) :-
	check_literals(H,FL),
	pr_oneof_rule(Id,H,H),
	pr_oneof(Id,T,FL).

pr_oneof_rule(_,[],_).
pr_oneof_rule(Id,[H|T],Fs) :-
	delete(Fs,H,Res),
	negate_list(Res,C),
	pr_positive(Id,H,Res),
	pr_static_law(Id,H,C),
	nl(Id),
	pr_oneof_rule(Id,T,Fs).

pr_positive(_,_,[]).
pr_positive(Id,X,[H|T]) :-
	pr_static_law(Id,neg(X),[H]),
	nl(Id),
	pr_positive(Id,X,T).

% pr_effects(+Id,+EfL,+AL,+FL): print rules encoding effects
% of non-sensing actions
pr_effects(_,[],_,_).
pr_effects(Id,[(A,L,C)|T],AL,FL) :-
	check_action(A,AL),
	check_literal(L,FL),
	check_literals(C,FL),
	make_atom(e,',T+1,P',L,H1),
	make_list(holds,',T,P',C,Bd1),
	make_atom(occ,',T,P',A,A1),
	append(['time(T)','path(P)'],[A1|Bd1],B1),
	pr_rule(Id,H1,B1),
	make_atom(pc,',T+1,P',L,H2),
	make_list(ph,',T,P',C,Bd2),
	append(['time(T)','path(P)'],[A1|Bd2],B2),
	pr_rule(Id,H2,B2),
	pr_effects(Id,T,AL,FL).

% pr_sense_effects(+Id,+EfL,+AL,+FL)
pr_sense_effects(_,[],_,_).
pr_sense_effects(Id,[(A,Ls)|T],AL,FL) :-
	check_action(A,AL),
	check_literals(Ls,FL),
	pr_sense_effect(Id,A,Ls),
	pr_sense_effects(Id,T,AL,FL).

pr_sense_effect(Id,A,[H|T]) :-
	pr_sense_effect1(Id,A,H),
	pr_sense_effect2(Id,A,T),
	pr_sense_effect3(Id,A,[H|T]).

pr_sense_effect1(Id,A,L) :-
	make_atom(br,',T,P,P',L,H),
	make_atom(occ,',T,P',A,A1),
	pr_rule(Id,H,['time(T)','path(P)',A1]).

pr_sense_effect2(_,_,[]).
pr_sense_effect2(Id,A,[H|T]) :-
	name('1{br(',L1),
	write_to_chars(H,L2),
	name(',T,P,P1):new_br(P,P1)}1',L3),
	append(L1,L2,L4),
	append(L4,L3,L5),
	name(Hd,L5),
	make_atom(occ,',T,P',A,A1),
	pr_rule(Id,Hd,['time(T)','path(P)',A1]),
	pr_sense_effect2(Id,A,T).

pr_sense_effect3(_,_,[]).
pr_sense_effect3(Id,A,[H|T]) :-
	make_atom(occ,',T,P',A,A1),
	make_atom(holds,',T,P',H,A2),
	pr_rule(Id,'',['time(T)','path(P)',A1,A2]),
	pr_sense_effect3(Id,A,T).

pr_hide(Id) :-
	write(Id,'hide time(T).\n'),
	write(Id,'hide time1(T).\n'),
	write(Id,'hide path(P).\n'),
	write(Id,'hide fluent(F).\n'),
	write(Id,'hide sense(F).\n'),
	write(Id,'hide literal(L).\n'),
	write(Id,'hide contrary(L1,L2).\n'),
	write(Id,'hide action(A).\n'),
	write(Id,'hide poss(A,T,P).\n'),
	write(Id,'hide holds(L,T,P).\n'),
	write(Id,'hide ph(L,T,P).\n'),
	write(Id,'hide e(L,T,P).\n'),
	write(Id,'hide pc(L,T,P).\n'),
	write(Id,'hide new_br(P,P1).\n'),
	write(Id,'hide used(T,P).\n'),
	write(Id,'hide goal(T,P).\n').
